@@ -46,10 +46,7 @@ GetOptions(\%args, qw(
     static has-libtommath has-libatomic_ops
     has-dyncall has-libffi pkgconfig=s
     build=s host=s big-endian jit! enable-jit
-    prefix=s bindir=s libdir=s mastdir=s
-    relocatable make-install asan ubsan
-    valgrind telemeh show-autovect
-    show-autovect-failed:s),
+    prefix=s bindir=s libdir=s mastdir=s make-install asan ubsan valgrind telemeh show-autovect show-autovect-failed:s),
 
     'no-optimize|nooptimize' => sub { $args{optimize} = 0 },
     'no-debug|nodebug' => sub { $args{debug} = 0 },
@@ -71,7 +68,7 @@ Please specify another installation target by using --prefix=PATH.
 ENOTTOCWD
 }
 
-# Override default target directories with command line arguments
+# Override default target directories with command line argumets
 my @target_dirs = qw{bindir libdir mastdir};
 for my $target (@target_dirs) {
     $config{$target} = $args{$target} if $args{$target};
@@ -94,11 +91,6 @@ if (-d '.git') {
 $args{optimize}     = 3 if not defined $args{optimize} or $args{optimize} eq "";
 $args{debug}        = 3 if defined $args{debug} and $args{debug} eq "";
 
-# Relocatability is not supported on AIX and OpenBSD.
-if ( $args{relocatable} && ($^O eq 'aix' || $^O eq 'openbsd') ) {
-    hardfail('Relocatability is not supported on ' . $^O .
-    ".\n    Leave off the --relocatable flag to do a non-relocatable build.");
-}
 
 for (qw(coverage instrument static big-endian has-libtommath has-sha has-libuv
         has-libatomic_ops asan ubsan valgrind show-vec)) {
@@ -178,17 +170,10 @@ $config{as}           = $config{cc} unless defined $config{as};
 # If we're in macOS, let's verify that the toolchain is consistent.
 if ($^O eq 'darwin') {
     my $gnu_count = 0;
-    my $gnu_toolchain = (exists $args{toolchain}) && ($args{toolchain} eq 'gnu');
-
-    unless ($gnu_toolchain) {
-        # When XCode toolchain is used then force use of XCode's make if
-        # available.
-        $config{make} = '/usr/bin/make' if -x '/usr/bin/make'; 
-    }
 
     # Here are the tools that seem to cause trouble.
     # If you see other ones, please add them to this list.
-    my @check_tools = qw/ar cc ld/;
+    my @check_tools = qw/ar cc ld make/;
     for my $tool (map { `which $_` } @config{@check_tools}) {
         chomp $tool;
         system "grep -b 'gnu' '$tool'"; # Apple utilities don't match `gnu`
@@ -198,14 +183,20 @@ if ($^O eq 'darwin') {
     }
 
     ## For a GNU toolchain, make sure that they're all GNU.
-    if ($gnu_toolchain && $gnu_count != scalar @check_tools) {
+    if (exists $args{toolchain} &&
+        $args{toolchain} eq 'gnu' &&
+        $gnu_count != scalar @check_tools)
+    {
         print "\nNot all tools in the toolchain are GNU. Please correct this and retry.\n"
             . "See README.markdown for more details.\n\n";
         exit -1;
     }
 
     ## Otherwise, make sure that none of them are GNU
-    elsif (!$gnu_toolchain && $gnu_count != 0) {
+    elsif ((!exists $args{toolchain} ||
+        $args{toolchain} ne 'gnu' ) &&
+        $gnu_count != 0)
+    {
         print "\nGNU tools detected, despite this not being a GNU-oriented build.\n"
             ." Please correct this and retry. See README.markdown for more details.\n\n";
         exit -1;
@@ -223,11 +214,6 @@ if ($config{cc} eq 'gcc' && !$config{can_specific_werror}) {
     $config{ccmiscflags} =~ s/^ +$//;
 }
 
-# Disable RETGUARD on OpenBSD, since it breaks the legojit.
-if ($^O eq 'openbsd' && $args{jit} && $config{cc} eq 'clang') {
-	$config{ccmiscflags} .= ' -fno-ret-protector';
-}
-
 # Set the remaining ldmiscflags. Do this after probing for gcc -Werror probe to not miss that change for the linker.
 $config{ldmiscflags}  = $config{ccmiscflags} unless defined $config{ldmiscflags};
 
@@ -431,23 +417,13 @@ $config{cflags} = join ' ', uniq(@cflags);
 my @ldflags = ($config{ldmiscflags});
 push @ldflags, $config{ldoptiflags}  if $args{optimize};
 push @ldflags, $config{lddebugflags} if $args{debug};
-push @ldflags, $config{ldinstflags}  if $args{instrument};
+push @ldflags, $config{ldinstflags}       if $args{instrument};
 push @ldflags, $config{ld_covflags}  if $args{coverage};
-if (not $args{static} and $config{prefix} ne '/usr') {
-    push @ldflags, $config{ldrpath_relocatable} if  $args{relocatable};
-    push @ldflags, $config{ldrpath}             if !$args{relocatable};
-}
-push @ldflags, '-fsanitize=address'  if $args{asan};
-push @ldflags, $ENV{LDFLAGS}         if $ENV{LDFLAGS};
+push @ldflags, $config{ldrpath}           if not $args{static} and $config{prefix} ne '/usr';
+push @ldflags, '-fsanitize=address' if $args{asan};
+push @ldflags, $ENV{LDFLAGS}  if $ENV{LDFLAGS};
 $config{ldflags} = join ' ', @ldflags;
 
-$config{moarshared} = '';
-# Switch shared lib compiler flags in relocatable case.
-if (not $args{static} and $config{prefix} ne '/usr') {
-    $config{moarshared} = $config{moarshared_relocatable}   if  $args{relocatable};
-    $config{moarshared} = $config{moarshared_norelocatable} if !$args{relocatable};
-}
-
 # setup library names
 $config{moarlib} = sprintf $config{lib}, $NAME;
 $config{moardll} = sprintf $config{dll}, $NAME;
@@ -527,7 +503,6 @@ if ($config{cc} eq 'cl') {
 build::probe::C_type_bool(\%config, \%defaults);
 build::probe::computed_goto(\%config, \%defaults);
 build::probe::pthread_yield(\%config, \%defaults);
-build::probe::check_fn_malloc_trim(\%config, \%defaults);
 if ($^O eq 'aix') {
     build::probe::numbits(\%config, \%defaults);
     $config{ldflags} = join(',', $config{ldflags}, '-bmaxdata:0x80000000')
@@ -937,7 +912,7 @@ __END__
                    [--toolchain <toolchain>] [--compiler <compiler>]
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
                    [--debug] [--optimize] [--instrument]
-                   [--static] [--prefix <path>] [--relocatable]
+                   [--static] [--prefix]
                    [--has-libtommath] [--has-sha] [--has-libuv]
                    [--has-libatomic_ops]
                    [--asan] [--ubsan] [--no-jit]
@@ -946,8 +921,8 @@ __END__
     ./Configure.pl --build <build-triple> --host <host-triple>
                    [--ar <ar>] [--cc <cc>] [--ld <ld>] [--make <make>]
                    [--debug] [--optimize] [--instrument]
-                   [--static] [--prefix <path>] [--relocatable]
-                   [--big-endian] [--make-install]
+                   [--static] [--big-endian] [--prefix]
+                   [--make-install]
 
 =head2 Use of environment variables
 
@@ -1079,12 +1054,6 @@ builds, the byte order is auto-detected.
 Install files in subdirectory /bin, /lib and /include of the supplied path.
 The default prefix is "install" if this option is not passed.
 
-=item --relocatable
-
-Search for the libmoar library relative to the executable path.
-(On AIX and OpenBSD MoarVM can not be built relocatable, since both OS'
-miss the necessary mechanism to make this work.)
-
 =item --bindir
 
 Install executable files in the supplied path.  The default is
