@@ -20,13 +20,10 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
     MVM_gc_root_temp_push(tc, (MVMCollectable **)&mro);
 
     flat_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&flat_list);
 
     class_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&class_list);
 
     attr_map_list = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_array_type);
-    MVM_gc_root_temp_push(tc, (MVMCollectable **)&attr_map_list);
 
     /* Walk through the parents list. */
     while (mro_idx)
@@ -47,9 +44,7 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
             MVMObject *attr_map = NULL;
 
             if (MVM_iter_istrue(tc, attr_iter)) {
-                MVM_gc_root_temp_push(tc, (MVMCollectable **)&attr_iter);
                 attr_map = MVM_repr_alloc_init(tc, MVM_hll_current(tc)->slurpy_hash_type);
-                MVM_gc_root_temp_push(tc, (MVMCollectable **)&attr_map);
             }
 
             while (MVM_iter_istrue(tc, attr_iter)) {
@@ -74,15 +69,12 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
                 MVM_repr_push_o(tc, flat_list, attr);
             }
 
-            if (attr_map) {
-                MVM_gc_root_temp_pop_n(tc, 2);
-            }
-
             /* Add to class list and map list. */
             MVM_repr_push_o(tc, class_list, current_class);
             MVM_repr_push_o(tc, attr_map_list, attr_map);
         }
         else {
+            MVM_gc_allocate_gen2_default_clear(tc);
             MVM_exception_throw_adhoc(tc,
                 "CStruct representation does not support multiple inheritance");
         }
@@ -90,7 +82,7 @@ static MVMObject * index_mapping_and_flat_list(MVMThreadContext *tc, MVMObject *
         MVM_gc_root_temp_pop(tc); /* current_class */
     }
 
-    MVM_gc_root_temp_pop_n(tc, 5);
+    MVM_gc_root_temp_pop_n(tc, 2); /* mro, st */
 
     /* We can now form the name map. */
     num_classes = MVM_repr_elems(tc, class_list);
@@ -121,9 +113,9 @@ static MVMint32 round_up_to_multi(MVMint32 i, MVMint32 m) {
 static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_info, MVMCStructREPRData *repr_data, MVMSTable *st) {
     /* Compute index mapping table and get flat list of attributes. */
     MVMObject *flat_list;
-    MVMROOT(tc, st, {
-        flat_list = index_mapping_and_flat_list(tc, repr_info, repr_data, st);
-    });
+    MVM_gc_root_temp_push(tc, (MVMCollectable **)&st);
+    MVM_gc_root_temp_push(tc, (MVMCollectable **)&repr_info);
+    flat_list = index_mapping_and_flat_list(tc, repr_info, repr_data, st);
 
     /* If we have no attributes in the index mapping, then just the header. */
     if (repr_data->name_to_index_mapping[0].class_key == NULL) {
@@ -144,8 +136,10 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
         MVMint32 cur_init_slot    = 0;
         MVMint32 i;
 
-        if (info_alloc == 0)
+        if (info_alloc == 0) {
+            MVM_gc_allocate_gen2_default_clear(tc);
             MVM_exception_throw_adhoc(tc, "Class %s has no attributes, which is illegal with the CStruct representation.", MVM_6model_get_stable_debug_name(tc, st));
+        }
 
         /* Allocate location/offset arrays and GC mark info arrays. */
         repr_data->num_attributes      = num_attrs;
@@ -171,6 +165,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
             MVMint32   align = ALIGNOF(void *);
 
             if (num_dimensions > 1) {
+                MVM_gc_allocate_gen2_default_clear(tc);
                 MVM_exception_throw_adhoc(tc,
                     "Only one dimensions supported in CStruct attribute");
             }
@@ -224,16 +219,22 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                         repr_data->attribute_locations[i]  |= MVM_CSTRUCT_ATTR_INLINED;
 
                         if (num_dimensions > 0) {
-                            MVMint64 dim_one =  MVM_repr_at_pos_i(tc, dimensions, 0);
+                            MVMint64 dim_one     =  MVM_repr_at_pos_i(tc, dimensions, 0);
+                            MVMObject *elem_type = carray_repr_data->elem_type;
 
                             // How do we distinguish between these members:
                             // a) struct  foo [32] alias;
                             // b) struct *foo [32] alias;
                             if (carray_repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_CSTRUCT) {
-                                MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(carray_repr_data->elem_type)->REPR_data;
+                                MVMCStructREPRData *cstruct_repr_data = (MVMCStructREPRData *)STABLE(elem_type)->REPR_data;
                                 bits                                  = cstruct_repr_data->struct_size * 8 * dim_one;
                                 align                                 = cstruct_repr_data->struct_align;
                             }
+                            else if (carray_repr_data->elem_kind == MVM_CARRAY_ELEM_KIND_NUMERIC) {
+                                const MVMStorageSpec *spec = REPR(elem_type)->get_storage_spec(tc, STABLE(elem_type));
+                                bits  = bits * dim_one;
+                                align = spec->align;
+                            }
                             else {
                                 bits  = bits * dim_one;
                             }
@@ -283,6 +284,7 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                     MVM_ASSIGN_REF(tc, &(st->header), repr_data->member_types[i], type);
                 }
                 else {
+                    MVM_gc_allocate_gen2_default_clear(tc);
                     MVM_exception_throw_adhoc(tc,
                         "CStruct representation only handles attributes of type:\n"
                         "  (u)int8, (u)int16, (u)int32, (u)int64, (u)long, (u)longlong, num32, num64, (s)size_t, bool, Str\n"
@@ -290,11 +292,13 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
                 }
             }
             else {
+                MVM_gc_allocate_gen2_default_clear(tc);
                 MVM_exception_throw_adhoc(tc,
                     "CStruct representation requires the types of all attributes to be specified");
             }
 
             if (bits % 8) {
+                MVM_gc_allocate_gen2_default_clear(tc);
                  MVM_exception_throw_adhoc(tc,
                     "CStruct only supports native types that are a multiple of 8 bits wide (was passed: %"PRId32")", bits);
             }
@@ -322,6 +326,8 @@ static void compute_allocation_strategy(MVMThreadContext *tc, MVMObject *repr_in
         if (repr_data->initialize_slots)
             repr_data->initialize_slots[cur_init_slot] = -1;
     }
+
+    MVM_gc_root_temp_pop_n(tc, 2); /* repr_info, st */
 }
 
 /* Helper for reading a pointer at the specified offset. */
@@ -372,7 +378,9 @@ static void compose(MVMThreadContext *tc, MVMSTable *st, MVMObject *repr_info) {
     /* Compute allocation strategy. */
     MVMCStructREPRData *repr_data = MVM_calloc(1, sizeof(MVMCStructREPRData));
     MVMObject *attr_info = MVM_repr_at_key_o(tc, repr_info, tc->instance->str_consts.attribute);
+    MVM_gc_allocate_gen2_default_set(tc);
     compute_allocation_strategy(tc, attr_info, repr_data, st);
+    MVM_gc_allocate_gen2_default_clear(tc);
     st->REPR_data = repr_data;
 }
 
@@ -452,7 +460,11 @@ static void get_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                     if (cobj) {
                         MVMObject **child_objs = body->child_objs;
                         if (type == MVM_CSTRUCT_ATTR_CARRAY) {
-                            obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
+                            if (repr_data->attribute_locations[slot] & MVM_CSTRUCT_ATTR_INLINED)
+                                obj = MVM_nativecall_make_carray(tc, typeobj,
+                                    (char *)body->cstruct + repr_data->struct_offsets[slot]);
+                            else
+                                obj = MVM_nativecall_make_carray(tc, typeobj, cobj);
                         }
                         else if(type == MVM_CSTRUCT_ATTR_CSTRUCT) {
                             if (repr_data->attribute_locations[slot] & MVM_CSTRUCT_ATTR_INLINED)
@@ -566,7 +578,11 @@ static void bind_attribute(MVMThreadContext *tc, MVMSTable *st, MVMObject *root,
                         if (REPR(value)->ID != MVM_REPR_ID_MVMCArray)
                             MVM_exception_throw_adhoc(tc,
                                 "Can only store CArray attribute in CArray slot in CStruct");
-                        cobj = ((MVMCArray *)value)->body.storage;
+                        if (repr_data->attribute_locations[slot] & MVM_CSTRUCT_ATTR_INLINED)
+                            cobj = ((MVMCArray *)value)->body.storage
+                                 = (char *)body->cstruct + repr_data->struct_offsets[slot];
+                        else
+                            cobj = ((MVMCArray *)value)->body.storage;
                     }
                     else if (type == MVM_CSTRUCT_ATTR_CSTRUCT) {
                         if (REPR(value)->ID != MVM_REPR_ID_MVMCStruct)
@@ -831,6 +847,140 @@ static void deserialize_stable_size(MVMThreadContext *tc, MVMSTable *st, MVMSeri
     st->size = sizeof(MVMCStruct);
 }
 
+static void add_slot_name_comment(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins,
+                                  MVMString *name, MVMSpeshFacts *type_handle_facts, MVMSTable *st) {
+    if (MVM_spesh_debug_enabled(tc)) {
+        char *name_cstr = MVM_string_utf8_encode_C_string(tc, name);
+        if (type_handle_facts->type != st->WHAT) {
+            MVM_spesh_graph_add_comment(tc, g, ins, "%s of '%s' in %s of a %s",
+                    ins->info->name,
+                    name_cstr,
+                    MVM_6model_get_debug_name(tc, type_handle_facts->type),
+                    MVM_6model_get_stable_debug_name(tc, st));
+        }
+        else {
+            MVM_spesh_graph_add_comment(tc, g, ins, "%s of '%s' in %s",
+                    ins->info->name,
+                    name_cstr,
+                    MVM_6model_get_debug_name(tc, type_handle_facts->type));
+        }
+        MVM_free(name_cstr);
+    }
+}
+static MVMString * spesh_attr_name(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshOperand o, MVMint32 indirect) {
+    if (indirect) {
+        MVMSpeshFacts *name_facts = MVM_spesh_get_and_use_facts(tc, g, o);
+        if (name_facts->flags & MVM_SPESH_FACT_KNOWN_VALUE)
+            return name_facts->value.s;
+        else
+            return NULL;
+    }
+    else {
+        return MVM_spesh_get_string(tc, g, o);
+    }
+}
+static void make_deref_op(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins, MVMSpeshOperand orig, MVMSpeshOperand temp) {
+    MVMSpeshIns *deref_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
+    deref_ins->info = MVM_op_get_op(MVM_OP_sp_get_i64);
+    deref_ins->operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
+    deref_ins->operands[0] = temp;
+    deref_ins->operands[1] = orig;
+    deref_ins->operands[2].lit_i16 = offsetof(MVMCStruct, body.cstruct);
+
+    MVM_spesh_usages_add_by_reg(tc, g, orig, deref_ins);
+    MVM_spesh_get_facts(tc, g, deref_ins->operands[0])->writer = deref_ins;
+
+    MVM_spesh_manipulate_insert_ins(tc, bb, ins->prev, deref_ins);
+}
+static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpeshBB *bb, MVMSpeshIns *ins) {
+    MVMCStructREPRData *repr_data = (MVMCStructREPRData *)st->REPR_data;
+    MVMuint16             opcode    = ins->info->opcode;
+    /* Can only use sp_get_i64 to deref a pointer if we're on 64bit */
+    if (sizeof(MVMObject *) != sizeof(MVMuint64))
+        return;
+    if (!repr_data)
+        return;
+    switch (opcode) {
+    /*case MVM_OP_getattr_i:
+    case MVM_OP_getattrs_i: {
+        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
+        MVMSpeshFacts *ch_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[2]);
+        MVMString     *name     = spesh_attr_name(tc, g, ins->operands[3], opcode == MVM_OP_getattrs_i);
+        if (name && ch_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE && ch_facts->type
+            && obj_facts->flags & MVM_SPESH_FACT_CONCRETE) {
+            MVMint64 slot = try_get_slot(tc, repr_data, ch_facts->type, name);
+            if (slot >= 0 && repr_data->flattened_stables[slot]) {
+                MVMSTable      *flat_st = repr_data->flattened_stables[slot];
+                const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
+                add_slot_name_comment(tc, g, ins, name, ch_facts, st);
+                if (flat_st->REPR->ID == MVM_REPR_ID_P6int && (
+                            flat_ss->bits == 64 || flat_ss->bits == 32)) {
+                    MVMSpeshOperand temp_reg = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_int64);
+                    MVMSpeshOperand orig_target = ins->operands[1];
+                    fprintf(stderr, "speshed a %s on CStruct. please don't explode\n", ins->info->name);
+                    if (opcode == MVM_OP_getattrs_i)
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
+                    ins->info = MVM_op_get_op(
+                            flat_ss->bits == 64 ? MVM_OP_sp_get_i64
+                                                : MVM_OP_sp_get_i32);
+                    ins->operands[1] = temp_reg;
+                    ins->operands[2].lit_i16 = repr_data->struct_offsets[slot];
+
+                    MVM_spesh_usages_add_by_reg(tc, g, temp_reg, ins);
+
+                    make_deref_op(tc, g, bb, ins, orig_target, temp_reg);
+                }
+            }
+        }
+        break;
+    }*/
+    case MVM_OP_bindattr_i:
+    case MVM_OP_bindattrs_i: {
+        MVMSpeshFacts *obj_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[0]);
+        MVMSpeshFacts *ch_facts = MVM_spesh_get_and_use_facts(tc, g, ins->operands[1]);
+        MVMString     *name     = spesh_attr_name(tc, g, ins->operands[2], opcode == MVM_OP_getattrs_i);
+        if (name && ch_facts->flags & MVM_SPESH_FACT_KNOWN_TYPE && ch_facts->type
+            && obj_facts->flags & MVM_SPESH_FACT_CONCRETE) {
+            MVMint64 slot = try_get_slot(tc, repr_data, ch_facts->type, name);
+            if (slot >= 0 && repr_data->flattened_stables[slot]) {
+                MVMSTable      *flat_st = repr_data->flattened_stables[slot];
+                const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
+                add_slot_name_comment(tc, g, ins, name, ch_facts, st);
+                if (flat_st->REPR->ID == MVM_REPR_ID_P6int && (
+                            flat_ss->bits == 64 || flat_ss->bits == 32 || flat_ss->bits == 16 || flat_ss->bits == 8)) {
+                    MVMSpeshOperand temp_reg = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_int64);
+                    MVMSpeshOperand orig_target = ins->operands[0];
+                    if (opcode == MVM_OP_bindattrs_i)
+                        MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[1], ins);
+                    MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[0], ins);
+                    ins->info = MVM_op_get_op(
+                            flat_ss->bits == 64 ? MVM_OP_sp_bind_i64
+                          : flat_ss->bits == 32 ? MVM_OP_sp_bind_i32
+                          : flat_ss->bits == 16 ? MVM_OP_sp_bind_i16
+                          :                       MVM_OP_sp_bind_i8);
+                    ins->operands[0] = temp_reg;
+                    ins->operands[1].lit_i16 = repr_data->struct_offsets[slot];
+                    ins->operands[2] = ins->operands[3];
+
+                    MVM_spesh_usages_add_by_reg(tc, g, temp_reg, ins);
+
+                    make_deref_op(tc, g, bb, ins, orig_target, temp_reg);
+
+                    MVM_spesh_manipulate_release_temp_reg(tc, g, temp_reg);
+                }
+            }
+        }
+        break;
+    }
+    default:
+        MVM_spesh_graph_add_comment(tc, g, ins, "%s unsupported in CStruct %s",
+                ins->info->name,
+                MVM_6model_get_stable_debug_name(tc, st));
+    }
+}
+
 /* Initializes the representation. */
 const MVMREPROps * MVMCStruct_initialize(MVMThreadContext *tc) {
     return &CStruct_this_repr;
@@ -865,7 +1015,7 @@ static const MVMREPROps CStruct_this_repr = {
     gc_mark_repr_data,
     gc_free_repr_data,
     compose,
-    NULL, /* spesh */
+    spesh, /* spesh */
     "CStruct", /* name */
     MVM_REPR_ID_MVMCStruct,
     NULL, /* unmanaged_size */
