@@ -1133,6 +1133,7 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
     MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)st->REPR_data;
     MVMuint16 num_attributes = repr_data->num_attributes;
     MVMuint16 i;
+    data = MVM_p6opaque_real_data(tc, data);
     for (i = 0; i < num_attributes; i++) {
         MVMuint16 a_offset = repr_data->attribute_offsets[i];
         MVMSTable *a_st = repr_data->flattened_stables[i];
@@ -1500,11 +1501,15 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
                 MVMSTable      *flat_st = repr_data->flattened_stables[slot];
                 const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
                 add_slot_name_comment(tc, g, ins, name, ch_facts, st);
-                if (flat_st->REPR->ID == MVM_REPR_ID_P6int && flat_ss->bits == 64) {
+                if (flat_st->REPR->ID == MVM_REPR_ID_P6int &&
+                        (flat_ss->bits == 64 || flat_ss->bits == 32)) {
                     if (opcode == MVM_OP_getattrs_i)
                         MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[3], ins);
                     MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
-                    ins->info = MVM_op_get_op(MVM_OP_sp_p6oget_i);
+                    ins->info = MVM_op_get_op(
+                            flat_ss->bits == 64 ? MVM_OP_sp_p6oget_i
+                                                : MVM_OP_sp_p6oget_i32
+                            );
                     ins->operands[2].lit_i16 = repr_data->attribute_offsets[slot];
                 }
             }
@@ -1587,12 +1592,16 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             if (slot >= 0 && repr_data->flattened_stables[slot]) {
                 MVMSTable      *flat_st = repr_data->flattened_stables[slot];
                 const MVMStorageSpec *flat_ss = flat_st->REPR->get_storage_spec(tc, flat_st);
-                if (flat_st->REPR->ID == MVM_REPR_ID_P6int && flat_ss->bits == 64) {
+                if (flat_st->REPR->ID == MVM_REPR_ID_P6int &&
+                        (flat_ss->bits == 64 || flat_ss->bits == 32)) {
                     add_slot_name_comment(tc, g, ins, name, ch_facts, st);
                     if (opcode == MVM_OP_bindattrs_i)
                         MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[2], ins);
                     MVM_spesh_usages_delete_by_reg(tc, g, ins->operands[1], ins);
-                    ins->info = MVM_op_get_op(MVM_OP_sp_p6obind_i);
+                    ins->info = MVM_op_get_op(
+                            flat_ss->bits == 64 ? MVM_OP_sp_p6obind_i
+                                                : MVM_OP_sp_p6obind_i32
+                            );
                     ins->operands[1].lit_i16 = repr_data->attribute_offsets[slot];
                     ins->operands[2] = ins->operands[3];
                 }
@@ -1782,6 +1791,10 @@ static void spesh(MVMThreadContext *tc, MVMSTable *st, MVMSpeshGraph *g, MVMSpes
             }
         }
         break;
+    default:
+        MVM_spesh_graph_add_comment(tc, g, ins, "reprop %s unsupported in P6Opaque %s",
+                ins->info->name,
+                MVM_6model_get_stable_debug_name(tc, st));
     }
 }
 
@@ -1879,6 +1892,16 @@ MVMuint16 MVM_p6opaque_get_bigint_offset(MVMThreadContext *tc, MVMSTable *st) {
     return 0;
 }
 
+/* Gets the attribute index given we know the slot offset. */
+MVMuint32 MVM_p6opaque_offset_to_attr_idx(MVMThreadContext *tc, MVMObject *type, size_t offset) {
+    MVMP6opaqueREPRData *repr_data = (MVMP6opaqueREPRData *)type->st->REPR_data;
+    MVMuint32 i;
+    for (i = 0; i < repr_data->num_attributes; i++)
+        if (repr_data->attribute_offsets[i] == offset)
+            return i;
+    MVM_oops(tc, "P6opaque: slot offset not found");
+}
+
 #ifdef DEBUG_HELPERS
 /* This is meant to be called in a debugging session and not used anywhere else.
  * Plese don't delete. */
@@ -1916,7 +1939,7 @@ static void dump_p6opaque(MVMThreadContext *tc, MVMObject *obj, int nested) {
                         MVMSTable * const attr_st = repr_data->flattened_stables[slot];
                         if (attr_st == NULL) {
                             MVMObject *value = get_obj_at_offset(data, offset);
-                            if (value != NULL) {
+                            if (value != NULL && REPR(value)->ID == MVM_REPR_ID_P6opaque) {
                                 fprintf(stderr, "=");
                                 dump_p6opaque(tc, value, 1);
                             }
@@ -1929,7 +1952,7 @@ static void dump_p6opaque(MVMThreadContext *tc, MVMObject *obj, int nested) {
                             }
                             else if (attr_st->REPR->ID == MVM_REPR_ID_P6int) {
                                 MVMint64 val = attr_st->REPR->box_funcs.get_int(tc, attr_st, obj, (char *)data + offset);
-                                fprintf(stderr, "=%ld", val);
+                                fprintf(stderr, "=%"PRIi64, val);
                             }
                             else {
                                 fprintf(stderr, "[%d]=%s", repr_data->attribute_offsets[slot], MVM_6model_get_stable_debug_name(tc, attr_st));
