@@ -404,7 +404,6 @@ void init_c_call_node(MVMThreadContext *tc, MVMSpeshGraph *sg, MVMJitNode *node,
         node->u.call.args = NULL;
     }
     node->u.call.num_args = num_args;
-    node->u.call.has_vargs = 0;
     node->u.call.rv_mode = MVM_JIT_RV_VOID;
     node->u.call.rv_idx  = -1;
 }
@@ -520,6 +519,12 @@ MVMJitGraph *MVM_nativecall_jit_graph_for_caller_code(
                         ? is_rw ? MVM_JIT_ARG_I64_RW : MVM_JIT_ARG_I64
                         : is_rw ? MVM_JIT_PARAM_I64_RW : MVM_JIT_PARAM_I64;
                     break;
+                case MVM_NATIVECALL_ARG_DOUBLE:
+                    if (is_rw) goto fail;
+                    arg_type = dst == -1
+                        ? MVM_JIT_ARG_DOUBLE
+                        : MVM_JIT_PARAM_DOUBLE;
+                    break;
                 case MVM_NATIVECALL_ARG_CPOINTER:
                     if (is_rw) goto fail;
                     arg_type = dst == -1 ? MVM_JIT_ARG_PTR : MVM_JIT_PARAM_PTR;
@@ -1008,13 +1013,13 @@ void MVM_nativecall_refresh(MVMThreadContext *tc, MVMObject *cthingy) {
         for (i = 0; i < repr_data->num_attributes; i++) {
             MVMint32 kind = repr_data->attribute_locations[i] & MVM_CSTRUCT_ATTR_MASK;
             MVMint32 slot = repr_data->attribute_locations[i] >> MVM_CSTRUCT_ATTR_SHIFT;
-            void *cptr = NULL;   /* The pointer in the C storage. */
+            void **cptr = NULL;  /* Address of the struct member holding the pointer in the C storage. */
             void *objptr = NULL; /* The pointer in the object representing the C object. */
 
             if (kind == MVM_CSTRUCT_ATTR_IN_STRUCT || !body->child_objs[slot])
                 continue;
 
-            cptr = (void*)((uintptr_t)storage + (uintptr_t)repr_data->struct_offsets[i]);
+            cptr = (void**)((uintptr_t)storage + (uintptr_t)repr_data->struct_offsets[i]);
             if (IS_CONCRETE(body->child_objs[slot])) {
                 switch (kind) {
                     case MVM_CSTRUCT_ATTR_CARRAY:
@@ -1045,7 +1050,7 @@ void MVM_nativecall_refresh(MVMThreadContext *tc, MVMObject *cthingy) {
                 objptr = NULL;
             }
 
-            if (objptr != cptr)
+            if (objptr != *cptr)
                 body->child_objs[slot] = NULL;
             else
                 MVM_nativecall_refresh(tc, body->child_objs[slot]);
