@@ -190,7 +190,7 @@ static void validate_branch_targets(Validator *val) {
 
         if ((flag & MVM_BC_branch_target) && !(flag & MVM_BC_op_boundary))
             fail(val, MSG(val, "branch targets offset %" PRIu32
-                "within instruction %" PRIu32), pos, instr);
+                " within instruction %" PRIu32), pos, instr);
     }
 }
 
@@ -292,8 +292,8 @@ static void validate_reg_operand(Validator *val, MVMuint32 flags) {
     }
 
     if (reg_type != operand_type)
-        fail(val, MSG(val, "operand type %"PRIu32" does not match register type %"PRIu32),
-                operand_type, reg_type);
+        fail(val, MSG(val, "operand type %" PRIu32 " does not match register type %" PRIu32 " for op %s in frame %s"),
+                operand_type, reg_type, val->cur_info->name, MVM_string_utf8_maybe_encode_C_string(val->tc, val->frame->body.name));
 
 next_operand:
     val->cur_op += 2;
@@ -468,6 +468,9 @@ static void validate_sequence(Validator *val) {
         type = val->cur_mark[0];
         id   = val->cur_mark[1];
 
+        if (val->cur_info->specializable)
+            val->frame->body.specializable = 1;
+
         switch (type) {
             case MARK_special:
                 if (id == seq_id)
@@ -621,6 +624,9 @@ static void validate_block(Validator *val) {
         type = val->cur_mark[0];
         id   = val->cur_mark[1];
 
+        if (val->cur_info->specializable)
+            val->frame->body.specializable = 1;
+
         if (id != block_id)
             fail(val, MSG(val, "expected instruction marked '%c' but got '%c'"),
                     block_id, id);
