@@ -57,23 +57,23 @@ int MVM_bigint_mp_set_uint64(mp_int * a, MVMuint64 b) {
  *  https://github.com/czurnieden/libtommath/blob/master/bn_mp_get_double.c
  *  and slightly modified to fit MoarVM's setup.
  */
-static const int mp_get_double_digits_needed
+static const int MVM_mp_get_double_digits_needed
 = ((MANTISSA_BITS_IN_DOUBLE + DIGIT_BIT) / DIGIT_BIT) + 1;
-static const double mp_get_double_multiplier = (double)(MP_MASK + 1);
+static const double MVM_mp_get_double_multiplier = (double)(MP_MASK + 1);
 
-static MVMnum64 mp_get_double(mp_int *a, int shift) {
+static MVMnum64 MVM_mp_get_double_shift(mp_int *a, int shift) {
     MVMnum64 d;
     int i, limit, final_shift;
     d = 0.0;
 
     mp_clamp(a);
     i = a->used;
-    limit = (i <= mp_get_double_digits_needed)
-        ? 0 : i - mp_get_double_digits_needed;
+    limit = (i <= MVM_mp_get_double_digits_needed)
+        ? 0 : i - MVM_mp_get_double_digits_needed;
 
     while (i-- > limit) {
         d += a->dp[i];
-        d *= mp_get_double_multiplier;
+        d *= MVM_mp_get_double_multiplier;
     }
 
     if (a->sign == MP_NEG)
@@ -377,6 +377,19 @@ MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MV
 }
 
 #define MVM_BIGINT_BINARY_OP_SIMPLE(opname, SMALLINT_OP) \
+void MVM_bigint_fallback_##opname(MVMThreadContext *tc, MVMP6bigintBody *ba, MVMP6bigintBody *bb, \
+                                  MVMP6bigintBody *bc) { \
+    mp_int *tmp[2] = { NULL, NULL }; \
+    mp_int *ia, *ib, *ic; \
+    ia = force_bigint(ba, tmp); \
+    ib = force_bigint(bb, tmp); \
+    ic = MVM_malloc(sizeof(mp_int)); \
+    mp_init(ic); \
+    mp_##opname(ia, ib, ic); \
+    store_bigint_result(bc, ic); \
+    clear_temp_bigints(tmp, 2); \
+    adjust_nursery(tc, bc); \
+} \
 MVMObject * MVM_bigint_##opname(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b) { \
     MVMP6bigintBody *ba, *bb, *bc; \
     MVMObject *result; \
@@ -634,7 +647,7 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
     } else {
         MVMint32 num   = ba->u.smallint.value;
         MVMint32 denom = bb->u.smallint.value;
-        MVMint32 value;
+        MVMint64 value;
         if ((cmp_a == MP_LT) ^ (cmp_b == MP_LT)) {
             if (denom == 0) {
                 MVM_exception_throw_adhoc(tc, "Division by zero");
@@ -645,7 +658,7 @@ MVMObject *MVM_bigint_div(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
                 value = num / denom;
             }
         } else {
-            value = num / denom;
+            value = (MVMint64)num / denom;
         }
         store_int64_result(bc, value);
     }
@@ -701,8 +714,8 @@ MVMObject * MVM_bigint_pow(MVMThreadContext *tc, MVMObject *a, MVMObject *b,
         }
     }
     else {
-        MVMnum64 f_base = mp_get_double(base, 0);
-        MVMnum64 f_exp = mp_get_double(exponent, 0);
+        MVMnum64 f_base = MVM_mp_get_double_shift(base, 0);
+        MVMnum64 f_exp = MVM_mp_get_double_shift(exponent, 0);
         r = MVM_repr_box_num(tc, num_type, pow(f_base, f_exp));
     }
     clear_temp_bigints(tmp, 2);
@@ -811,11 +824,12 @@ MVMObject *MVM_bigint_not(MVMThreadContext *tc, MVMObject *result_type, MVMObjec
     return result;
 }
 
-void MVM_bigint_expmod(MVMThreadContext *tc, MVMObject *result, MVMObject *a, MVMObject *b, MVMObject *c) {
+MVMObject *MVM_bigint_expmod(MVMThreadContext *tc, MVMObject *result_type, MVMObject *a, MVMObject *b, MVMObject *c) {
     MVMP6bigintBody *ba = get_bigint_body(tc, a);
     MVMP6bigintBody *bb = get_bigint_body(tc, b);
     MVMP6bigintBody *bc = get_bigint_body(tc, c);
-    MVMP6bigintBody *bd = get_bigint_body(tc, result);
+    MVMP6bigintBody *bd;
+    MVMObject       *result;
 
     mp_int *tmp[3] = { NULL, NULL, NULL };
 
@@ -825,10 +839,18 @@ void MVM_bigint_expmod(MVMThreadContext *tc, MVMObject *result, MVMObject *a, MV
     mp_int *id = MVM_malloc(sizeof(mp_int));
     mp_init(id);
 
+    MVMROOT3(tc, a, b, c, {
+        result = MVM_repr_alloc_init(tc, result_type);
+    });
+
+    bd = get_bigint_body(tc, result);
+
     mp_exptmod(ia, ib, ic, id);
     store_bigint_result(bd, id);
     clear_temp_bigints(tmp, 3);
     adjust_nursery(tc, bd);
+
+    return result;
 }
 
 void MVM_bigint_from_str(MVMThreadContext *tc, MVMObject *a, const char *buf) {
@@ -922,6 +944,78 @@ MVMObject * MVM_bigint_from_bigint(MVMThreadContext *tc, MVMObject *result_type,
     return result;
 }
 
+
+/* returns size of ASCII reprensentation */
+static int mp_faster_radix_size (mp_int *a, int radix, int *size)
+{
+    int     res, digs;
+    mp_int  t;
+    mp_digit d;
+
+    *size = 0;
+
+    /* make sure the radix is in range */
+    if ((radix < 2) || (radix > 64))
+        return MP_VAL;
+
+    if (mp_iszero(a) == MP_YES) {
+        *size = 2;
+        return MP_OKAY;
+    }
+
+    /* special case for binary */
+    if (radix == 2) {
+        *size = mp_count_bits(a) + ((a->sign == MP_NEG) ? 1 : 0) + 1;
+        return MP_OKAY;
+    }
+
+    digs = 0; /* digit count */
+
+    if (a->sign == MP_NEG)
+        ++digs;
+
+    /* init a copy of the input */
+    if ((res = mp_init_copy(&t, a)) != MP_OKAY)
+        return res;
+
+    /* force temp to positive */
+    t.sign = MP_ZPOS;
+
+    /* fetch out all of the digits */
+
+#if DIGIT_BIT == 60
+    /* Optimization for base-10 numbers.
+     * Logic is designed for 60-bit mp_digit, with 100000000000000000
+     * being the largest 10**n that can fit into it, which gives us 17 digits.
+     * So we reduce the number in 17 digit chunks, until we get to a number
+     * small enough to fit into a single mp_digit.
+     */
+    if (radix == 10) {
+        mp_clamp(&t);
+        while ((&t)->used > 1) {
+            if ((res = mp_div_d(&t, (mp_digit) 100000000000000000, &t, &d)) != MP_OKAY) {
+              mp_clear(&t);
+              return res;
+            }
+            digs += 17;
+        }
+    }
+#endif
+
+    while (mp_iszero(&t) == MP_NO) {
+        if ((res = mp_div_d(&t, (mp_digit) radix, &t, &d)) != MP_OKAY) {
+            mp_clear(&t);
+            return res;
+        }
+        ++digs;
+    }
+    mp_clear(&t);
+
+    /* return digs + 1, the 1 is for the NULL byte that would be required. */
+    *size = digs + 1;
+    return MP_OKAY;
+}
+
 MVMString * MVM_bigint_to_str(MVMThreadContext *tc, MVMObject *a, int base) {
     MVMP6bigintBody *body = get_bigint_body(tc, a);
     if (MVM_BIGINT_IS_BIG(body)) {
@@ -930,7 +1024,7 @@ MVMString * MVM_bigint_to_str(MVMThreadContext *tc, MVMObject *a, int base) {
         char *buf;
         MVMString *result;
         int is_malloced = 0;
-        mp_radix_size(i, base, &len);
+        mp_faster_radix_size(i, base, &len);
         if (len < 120) {
             buf = alloca(len);
         }
@@ -988,7 +1082,7 @@ MVMnum64 MVM_bigint_to_num(MVMThreadContext *tc, MVMObject *a) {
 
     if (MVM_BIGINT_IS_BIG(ba)) {
         mp_int *ia = ba->u.bigint;
-        return mp_get_double(ia, 0);
+        return MVM_mp_get_double_shift(ia, 0);
     } else {
         return (double)ba->u.smallint.value;
     }
@@ -1038,7 +1132,7 @@ MVMnum64 MVM_bigint_div_num(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
             if (mp_div(&scaled, ib, &scaled, NULL) != MP_OKAY)
                 MVM_exception_throw_adhoc(tc,
                     "Failed to preform bigint division");
-            c = mp_get_double(&scaled, bbits);
+            c = MVM_mp_get_double_shift(&scaled, bbits);
             mp_clear(&scaled);
         }
         clear_temp_bigints(tmp, 2);
@@ -1048,6 +1142,16 @@ MVMnum64 MVM_bigint_div_num(MVMThreadContext *tc, MVMObject *a, MVMObject *b) {
     return c;
 }
 
+
+/* 
+    The old version of LibTomMath has it publically defined the new one not,
+    so we can take the (non)existance as a marker.
+ */
+#ifndef MP_GEN_RANDOM_MAX
+#define MP_GEN_RANDOM_MAX MP_MASK
+#define MP_NEW_LTM_VERSION
+#endif
+
 MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b) {
     MVMObject *result;
     MVMP6bigintBody *ba;
@@ -1070,7 +1174,14 @@ MVMObject * MVM_bigint_rand(MVMThreadContext *tc, MVMObject *type, MVMObject *b)
 
     if (use_small_arithmetic) {
         if (MP_GEN_RANDOM_MAX >= abs(smallint_max)) {
-            mp_digit result_int = MP_GEN_RANDOM();
+            mp_digit result_int;
+#ifdef MP_NEW_LTM_VERSION
+            mp_digit p = MP_GEN_RANDOM_MAX;
+            mp_rand_digit(&p);
+            result_int = p;
+#else
+            result_int = MP_GEN_RANDOM();
+#endif
             result_int = result_int % smallint_max;
             if(have_to_negate)
                 result_int *= -1;
