@@ -143,6 +143,7 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "core/regionalloc.h"
 #include "spesh/dump.h"
 #include "spesh/debug.h"
+#include "spesh/pea.h"
 #include "spesh/graph.h"
 #include "spesh/codegen.h"
 #include "spesh/candidate.h"
@@ -195,7 +196,6 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "io/asyncsocket.h"
 #include "io/asyncsocketudp.h"
 #include "math/bigintops.h"
-#include "mast/driver.h"
 #include "core/intcache.h"
 #include "core/fixedsizealloc.h"
 #include "jit/graph.h"
@@ -204,7 +204,7 @@ MVM_PUBLIC const MVMint32 MVM_jit_support(void);
 #include "jit/register.h"
 #include "jit/tile.h"
 #include "jit/compile.h"
-#include "jit/log.h"
+#include "jit/dump.h"
 #include "jit/interface.h"
 #include "profiler/instrument.h"
 #include "profiler/log.h"
