@@ -202,6 +202,7 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
             case MVM_OP_getattrsref_i:
             case MVM_OP_getattrsref_n:
             case MVM_OP_getattrsref_s:
+            case MVM_OP_nativecallcast:
                 add_allocation_logging(tc, g, bb, ins);
                 break;
             case MVM_OP_nativecallinvoke:
@@ -309,6 +310,7 @@ typedef struct {
     MVMString *call_graph;
     MVMString *name;
     MVMString *id;
+    MVMString *parent;
     MVMString *file;
     MVMString *line;
     MVMString *entries;
@@ -321,6 +323,7 @@ typedef struct {
     MVMString *allocations;
     MVMString *spesh;
     MVMString *jit;
+    MVMString *replaced;
     MVMString *type;
     MVMString *count;
     MVMString *gcs;
@@ -331,8 +334,10 @@ typedef struct {
     MVMString *cleared_bytes;
     MVMString *retained_bytes;
     MVMString *promoted_bytes;
+    MVMString *promoted_bytes_unmanaged;
     MVMString *gen2_roots;
     MVMString *start_time;
+    MVMString *first_entry_time;
     MVMString *osr;
     MVMString *deopt_one;
     MVMString *deopt_all;
@@ -412,6 +417,7 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
                                         const MVMProfileCallNode *pcn, MVMObject *types_array) {
     MVMObject *node_hash  = new_hash(tc);
     MVMuint32  i;
+    MVMuint64 absolute_start_time;
 
     /* Let's see if we're dealing with a native call or a regular moar call */
     if (pcn->sf) {
@@ -440,7 +446,7 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
 
         /* Use static frame memory address to get a unique ID. */
         MVM_repr_bind_key_o(tc, node_hash, pds->id,
-            box_i(tc, (MVMint64)pcn->sf));
+            box_i(tc, (MVMint64)(uintptr_t)pcn->sf));
     } else {
         MVMString *function_name_string =
             MVM_string_utf8_c8_decode(tc, tc->instance->VMString,
@@ -456,7 +462,7 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
 
         /* Use the address of the name string as unique ID. a hack, but oh well. */
         MVM_repr_bind_key_o(tc, node_hash, pds->id,
-            box_i(tc, (MVMint64)pcn->native_target_name));
+            box_i(tc, (MVMint64)(uintptr_t)pcn->native_target_name));
     }
 
     /* Entry counts. */
@@ -473,10 +479,17 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
         MVM_repr_bind_key_o(tc, node_hash, pds->inlined_entries,
             box_i(tc, pcn->inlined_entries));
 
+    /* Use the main thread's start time for absolute timings */
+    absolute_start_time = tc->instance->main_thread->prof_data->start_time;
+
     /* Total (inclusive) time. */
     MVM_repr_bind_key_o(tc, node_hash, pds->inclusive_time,
         box_i(tc, pcn->total_time / 1000));
 
+    /* Earliest entry time */
+    MVM_repr_bind_key_o(tc, node_hash, pds->first_entry_time,
+        box_i(tc, (pcn->first_entry_time - absolute_start_time) / 1000));
+
     /* OSR and deopt counts. */
     if (pcn->osr_count)
         MVM_repr_bind_key_o(tc, node_hash, pds->osr,
@@ -497,7 +510,7 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
             MVMProfileAllocationCount *alloc = &pcn->alloc[i];
 
             MVMObject *type       = pcn->alloc[i].type;
-            MVMObject *type_info  = insert_if_not_exists(tc, pds, types_array, (MVMint64)type);
+            MVMObject *type_info  = insert_if_not_exists(tc, pds, types_array, (MVMint64)(uintptr_t)type);
 
             if (type_info) {
                 bind_extra_info(tc, type_info, pds->managed_size, box_i(tc, STABLE(type)->size));
@@ -507,7 +520,7 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
                 bind_extra_info(tc, type_info, pds->repr, box_s(tc, str(tc, REPR(type)->name)));
             }
 
-            MVM_repr_bind_key_o(tc, alloc_info, pds->id, box_i(tc, (MVMint64)type));
+            MVM_repr_bind_key_o(tc, alloc_info, pds->id, box_i(tc, (MVMint64)(uintptr_t)type));
             if (alloc->allocations_spesh)
                 MVM_repr_bind_key_o(tc, alloc_info, pds->spesh,
                     box_i(tc, alloc->allocations_spesh));
@@ -518,6 +531,9 @@ static MVMObject * dump_call_graph_node(MVMThreadContext *tc, ProfDumpStrs *pds,
                 box_i(tc, alloc->allocations_interp
                           + alloc->allocations_spesh
                           + alloc->allocations_jit));
+            if (alloc->scalar_replaced)
+                MVM_repr_bind_key_o(tc, alloc_info, pds->replaced,
+                    box_i(tc, alloc->scalar_replaced));
             MVM_repr_push_o(tc, alloc_list, alloc_info);
         }
     }
@@ -548,6 +564,10 @@ static MVMObject * dump_thread_data(MVMThreadContext *tc, ProfDumpStrs *pds,
     MVM_repr_bind_key_o(tc, thread_hash, pds->total_time,
         box_i(tc, (ptd->end_time - ptd->start_time) / 1000));
 
+    /* Add start time */
+    MVM_repr_bind_key_o(tc, thread_hash, pds->start_time,
+        box_i(tc, (ptd->start_time - absolute_start_time) / 1000));
+
     /* Add call graph. */
     if (ptd->call_graph)
         MVM_repr_bind_key_o(tc, thread_hash, pds->call_graph,
@@ -570,6 +590,8 @@ static MVMObject * dump_thread_data(MVMThreadContext *tc, ProfDumpStrs *pds,
             box_i(tc, ptd->gcs[i].retained_bytes));
         MVM_repr_bind_key_o(tc, gc_hash, pds->promoted_bytes,
             box_i(tc, ptd->gcs[i].promoted_bytes));
+        MVM_repr_bind_key_o(tc, gc_hash, pds->promoted_bytes_unmanaged,
+            box_i(tc, ptd->gcs[i].promoted_unmanaged_bytes));
         MVM_repr_bind_key_o(tc, gc_hash, pds->gen2_roots,
             box_i(tc, ptd->gcs[i].num_gen2roots));
         MVM_repr_bind_key_o(tc, gc_hash, pds->start_time,
@@ -586,6 +608,10 @@ static MVMObject * dump_thread_data(MVMThreadContext *tc, ProfDumpStrs *pds,
     MVM_repr_bind_key_o(tc, thread_hash, pds->thread,
         box_i(tc, othertc->thread_id));
 
+    /* Add parent thread id. */
+    MVM_repr_bind_key_o(tc, thread_hash, pds->parent,
+        box_i(tc, ptd->parent_thread_id));
+
     return thread_hash;
 }
 
@@ -607,6 +633,7 @@ void MVM_profile_dump_instrumented_data(MVMThreadContext *tc) {
         pds.call_graph      = str(tc, "call_graph");
         pds.name            = str(tc, "name");
         pds.id              = str(tc, "id");
+        pds.parent          = str(tc, "parent");
         pds.file            = str(tc, "file");
         pds.line            = str(tc, "line");
         pds.entries         = str(tc, "entries");
@@ -621,6 +648,7 @@ void MVM_profile_dump_instrumented_data(MVMThreadContext *tc) {
         pds.count           = str(tc, "count");
         pds.spesh           = str(tc, "spesh");
         pds.jit             = str(tc, "jit");
+        pds.replaced        = str(tc, "replaced");
         pds.gcs             = str(tc, "gcs");
         pds.time            = str(tc, "time");
         pds.full            = str(tc, "full");
@@ -631,6 +659,7 @@ void MVM_profile_dump_instrumented_data(MVMThreadContext *tc) {
         pds.promoted_bytes  = str(tc, "promoted_bytes");
         pds.gen2_roots      = str(tc, "gen2_roots");
         pds.start_time      = str(tc, "start_time");
+        pds.first_entry_time= str(tc, "first_entry_time");
         pds.osr             = str(tc, "osr");
         pds.deopt_one       = str(tc, "deopt_one");
         pds.deopt_all       = str(tc, "deopt_all");
@@ -642,6 +671,8 @@ void MVM_profile_dump_instrumented_data(MVMThreadContext *tc) {
         pds.has_unmanaged_data = str(tc, "has_unmanaged_data");
         pds.repr               = str(tc, "repr");
 
+        pds.promoted_bytes_unmanaged  = str(tc, "promoted_bytes_unmanaged");
+
         types_array = new_array(tc);
 
         MVM_repr_push_o(tc, tc->prof_data->collected_data, types_array);
