@@ -21,6 +21,10 @@ typedef struct {
 
     /* Copied frame handlers (which we'll update offsets of). */
     MVMFrameHandler *handlers;
+
+    /* Persisted deopt usage info, so we can recover it should be produce an
+     * inline from this bytecode in the future. */
+    MVM_VECTOR_DECL(MVMint32, deopt_usage_info);
 } SpeshWriterState;
 
 /* Write functions; all native endian. */
@@ -67,11 +71,10 @@ static void write_instructions(MVMThreadContext *tc, MVMSpeshGraph *g, SpeshWrit
     while (ins) {
         MVMint32 i;
 
-        /* Process any annotations. */
-        MVMSpeshAnn *ann              = ins->annotations;
-        MVMSpeshAnn *deopt_one_ann    = NULL;
-        MVMSpeshAnn *deopt_all_ann    = NULL;
-        MVMSpeshAnn *deopt_inline_ann = NULL;
+        /* Process any pre-annotations, and note if there were any deopt ones;
+         * if so, these should be processed later. */
+        MVMSpeshAnn *ann = ins->annotations;
+        MVMuint32 has_deopts = 0;
         while (ann) {
             switch (ann->type) {
             case MVM_SPESH_ANN_FH_START:
@@ -87,10 +90,9 @@ static void write_instructions(MVMThreadContext *tc, MVMSpeshGraph *g, SpeshWrit
                     ws->bytecode_pos;
                 break;
             case MVM_SPESH_ANN_DEOPT_ONE_INS:
-                deopt_one_ann = ann;
-                break;
             case MVM_SPESH_ANN_DEOPT_ALL_INS:
-                deopt_all_ann = ann;
+            case MVM_SPESH_ANN_DEOPT_INLINE:
+                has_deopts = 1;
                 break;
             case MVM_SPESH_ANN_INLINE_START:
                 g->inlines[ann->data.inline_idx].start = ws->bytecode_pos;
@@ -98,9 +100,6 @@ static void write_instructions(MVMThreadContext *tc, MVMSpeshGraph *g, SpeshWrit
             case MVM_SPESH_ANN_INLINE_END:
                 g->inlines[ann->data.inline_idx].end = ws->bytecode_pos;
                 break;
-            case MVM_SPESH_ANN_DEOPT_INLINE:
-                deopt_inline_ann = ann;
-                break;
             case MVM_SPESH_ANN_DEOPT_OSR:
                 g->deopt_addrs[2 * ann->data.deopt_idx + 1] = ws->bytecode_pos;
                 break;
@@ -109,7 +108,29 @@ static void write_instructions(MVMThreadContext *tc, MVMSpeshGraph *g, SpeshWrit
         }
 
         if (ins->info->opcode != MVM_SSA_PHI) {
-            /* Real instruction, not a phi. Emit opcode. */
+            /* Real instruction, not a phi. See if we need to save any deopt
+             * usage information at this location. */
+            if (g->facts && ins->info->num_operands >= 1 &&
+                    (ins->info->operands[0] & MVM_operand_rw_mask) == MVM_operand_write_reg) {
+                MVMSpeshDeoptUseEntry *deopt_users = MVM_spesh_get_facts(tc, g,
+                        ins->operands[0])->usage.deopt_users;
+                if (deopt_users) {
+                    MVMint32 count = 0;
+                    MVMSpeshDeoptUseEntry *count_entry = deopt_users;
+                    while (count_entry) {
+                        count++;
+                        count_entry = count_entry->next;
+                    }
+                    MVM_VECTOR_PUSH(ws->deopt_usage_info, ws->bytecode_pos);
+                    MVM_VECTOR_PUSH(ws->deopt_usage_info, count);
+                    while (deopt_users) {
+                        MVM_VECTOR_PUSH(ws->deopt_usage_info, deopt_users->deopt_idx);
+                        deopt_users = deopt_users->next;
+                    }
+                }
+            }
+
+            /* Emit opcode. */
             if (ins->info->opcode == (MVMuint16)-1) {
                 /* Ext op; resolve. */
                 MVMExtOpRecord *extops     = g->sf->body.cu->body.extops;
@@ -156,6 +177,9 @@ static void write_instructions(MVMThreadContext *tc, MVMSpeshGraph *g, SpeshWrit
                     case MVM_operand_int32:
                         write_int32(ws, ins->operands[i].lit_i32);
                         break;
+                    case MVM_operand_uint32:
+                        write_int32(ws, ins->operands[i].lit_ui32);
+                        break;
                     case MVM_operand_int64:
                         write_int64(ws, ins->operands[i].lit_i64);
                         break;
@@ -175,7 +199,11 @@ static void write_instructions(MVMThreadContext *tc, MVMSpeshGraph *g, SpeshWrit
                         write_int32(ws, ins->operands[i].lit_str_idx);
                         break;
                     case MVM_operand_ins: {
-                        MVMint32 offset = ws->bb_offsets[ins->operands[i].ins_bb->idx];
+                        MVMint32 bb_idx = ins->operands[i].ins_bb->idx;
+                        MVMint32 offset;
+                        if (bb_idx >= g->num_bbs)
+                            MVM_panic(1, "Spesh codegen: out of range BB index %d", bb_idx);
+                        offset = ws->bb_offsets[bb_idx];
                         if (offset >= 0) {
                             /* Already know where it is, so just write it. */
                             write_int32(ws, offset);
@@ -212,13 +240,20 @@ static void write_instructions(MVMThreadContext *tc, MVMSpeshGraph *g, SpeshWrit
             }
         }
 
-        /* If there was a deopt point annotation, update table. */
-        if (deopt_one_ann)
-            g->deopt_addrs[2 * deopt_one_ann->data.deopt_idx + 1] = ws->bytecode_pos;
-        if (deopt_all_ann)
-            g->deopt_addrs[2 * deopt_all_ann->data.deopt_idx + 1] = ws->bytecode_pos;
-        if (deopt_inline_ann)
-            g->deopt_addrs[2 * deopt_inline_ann->data.deopt_idx + 1] = ws->bytecode_pos;
+        /* If there were deopt point annotations, update table. */
+        if (has_deopts) {
+            ann = ins->annotations;
+            while (ann) {
+                switch (ann->type) {
+                case MVM_SPESH_ANN_DEOPT_ONE_INS:
+                case MVM_SPESH_ANN_DEOPT_ALL_INS:
+                case MVM_SPESH_ANN_DEOPT_INLINE:
+                    g->deopt_addrs[2 * ann->data.deopt_idx + 1] = ws->bytecode_pos;;
+                    break;
+                }
+                ann = ann->next;
+            }
+        }
 
         ins = ins->next;
     }
@@ -242,6 +277,7 @@ MVMSpeshCode * MVM_spesh_codegen(MVMThreadContext *tc, MVMSpeshGraph *g) {
     ws->fixup_bbs       = MVM_malloc(ws->alloc_fixups * sizeof(MVMSpeshBB *));
     for (i = 0; i < g->num_bbs; i++)
         ws->bb_offsets[i] = -1;
+    MVM_VECTOR_INIT(ws->deopt_usage_info, 0);
 
     /* Create copy of handlers, and -1 all offsets so we can catch missing
      * updates. */
@@ -249,7 +285,7 @@ MVMSpeshCode * MVM_spesh_codegen(MVMThreadContext *tc, MVMSpeshGraph *g) {
     if (hanlen) {
         ws->handlers = MVM_malloc(hanlen);
         memcpy(ws->handlers, g->handlers, hanlen);
-        for (i = 0; i < g->sf->body.num_handlers; i++) {
+        for (i = 0; i < g->num_handlers; i++) {
             ws->handlers[i].start_offset = -1;
             ws->handlers[i].end_offset   = -1;
             ws->handlers[i].goto_offset  = -1;
@@ -282,27 +318,52 @@ MVMSpeshCode * MVM_spesh_codegen(MVMThreadContext *tc, MVMSpeshGraph *g) {
         *((MVMuint32 *)(ws->bytecode + ws->fixup_locations[i])) =
             ws->bb_offsets[ws->fixup_bbs[i]->idx];
 
-    /* Ensure all handlers got fixed up. */
-    for (i = 0; i < g->sf->body.num_handlers; i++) {
-        if (ws->handlers[i].start_offset == -1 ||
-            ws->handlers[i].end_offset   == -1 ||
-            ws->handlers[i].goto_offset  == -1)
-            MVM_oops(tc, "Spesh: failed to fix up handlers (%d, %d, %d)",
+    /* Ensure all handlers that are reachable got fixed up. */
+    for (i = 0; i < g->num_handlers; i++) {
+        if (g->unreachable_handlers && g->unreachable_handlers[i]) {
+            ws->handlers[i].start_offset = -1;
+            ws->handlers[i].end_offset = -1;
+            ws->handlers[i].goto_offset = -1;
+        }
+        else if (ws->handlers[i].start_offset == -1 ||
+                 ws->handlers[i].end_offset   == -1 ||
+                 ws->handlers[i].goto_offset  == -1) {
+            MVM_oops(tc, "Spesh: failed to fix up handler %d in %s (%d, %d, %d)",
+                i,
+                MVM_string_utf8_maybe_encode_C_string(tc, g->sf->body.name),
                 (int)ws->handlers[i].start_offset,
                 (int)ws->handlers[i].end_offset,
                 (int)ws->handlers[i].goto_offset);
+        }
     }
 
     /* Ensure all inlines got fixed up. */
-    for (i = 0; i < g->num_inlines; i++)
-        if (g->inlines[i].start == -1 || g->inlines[i].end == -1)
-            MVM_oops(tc, "Spesh: failed to fix up inline %d", i);
+    for (i = 0; i < g->num_inlines; i++) {
+        if (g->inlines[i].unreachable) {
+            g->inlines[i].start = -1;
+            g->inlines[i].end = -1;
+        }
+        else {
+            if (g->inlines[i].start == -1 || g->inlines[i].end == -1)
+                MVM_oops(tc, "Spesh: failed to fix up inline %d (%s) %d %d",
+                    i,
+                    MVM_string_utf8_maybe_encode_C_string(tc, g->inlines[i].sf->body.name),
+                    g->inlines[i].start,
+                    g->inlines[i].end
+                );
+        }
+    }
+
+    /* Add terminating -1 to the deopt usage info. */
+    if (g->facts)
+        MVM_VECTOR_PUSH(ws->deopt_usage_info, -1);
 
     /* Produce result data structure. */
-    res                = MVM_malloc(sizeof(MVMSpeshCode));
-    res->bytecode      = ws->bytecode;
-    res->bytecode_size = ws->bytecode_pos;
-    res->handlers      = ws->handlers;
+    res                   = MVM_malloc(sizeof(MVMSpeshCode));
+    res->bytecode         = ws->bytecode;
+    res->bytecode_size    = ws->bytecode_pos;
+    res->handlers         = ws->handlers;
+    res->deopt_usage_info = ws->deopt_usage_info;
 
     /* Cleanup. */
     MVM_free(ws->bb_offsets);
