@@ -210,6 +210,14 @@ static void trunc_i16_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns
         tgt_facts->flags |= MVM_SPESH_FACT_KNOWN_VALUE;
     }
 }
+static void coerce_iu_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins) {
+    MVMSpeshFacts *src_facts = &g->facts[ins->operands[1].reg.orig][ins->operands[1].reg.i];
+    if (src_facts->flags & MVM_SPESH_FACT_KNOWN_VALUE) {
+        MVMSpeshFacts *tgt_facts = &g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i];
+        tgt_facts->value.i = (MVMint16)src_facts->value.i;
+        tgt_facts->flags |= MVM_SPESH_FACT_KNOWN_VALUE;
+    }
+}
 
 /* Discover facts from extops. */
 static void discover_extop(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *ins) {
@@ -393,9 +401,14 @@ static void log_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
         deopt_one_ann->next = NULL;
         guard->annotations = deopt_one_ann;
 
-        /* Move deopt usages to the preguard register. */
-        pre_facts->usage.deopt_users = facts->usage.deopt_users;
-        facts->usage.deopt_users = NULL;
+        /* Copy deopt usages to the preguard register. */
+        {
+            MVMSpeshDeoptUseEntry *due = facts->usage.deopt_users;
+            while (due) {
+                MVM_spesh_usages_add_deopt_usage(tc, g, pre_facts, due->deopt_idx);
+                due = due->next;
+            }
+        }
 
         /* Add entry in log guards table, and mark facts as depending on it. */
         if (g->num_log_guards % 16 == 0) {
@@ -480,6 +493,13 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 target_facts->flags |= MVM_SPESH_FACT_KNOWN_BOX_SRC;
                 break;
             }
+        case MVM_OP_coerce_sI: {
+                MVMSpeshFacts *target_facts = &(g->facts[ins->operands[0].reg.orig][ins->operands[0].reg.i]);
+                create_facts(tc, g,
+                    ins->operands[0].reg.orig, ins->operands[0].reg.i,
+                    ins->operands[2].reg.orig, ins->operands[2].reg.i);
+                break;
+            }
         case MVM_OP_add_I:
         case MVM_OP_sub_I:
         case MVM_OP_mul_I:
@@ -495,9 +515,14 @@ static void add_bb_facts(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
                 ins->operands[2].reg.orig, ins->operands[2].reg.i);
             break;
+        case MVM_OP_trunc_u16:
         case MVM_OP_trunc_i16:
             trunc_i16_facts(tc, g, ins);
             break;
+        case MVM_OP_coerce_ui:
+        case MVM_OP_coerce_iu:
+            trunc_i16_facts(tc, g, ins);
+            break;
         case MVM_OP_bootint:
             object_facts(tc, g,
                 ins->operands[0].reg.orig, ins->operands[0].reg.i,
