@@ -53,6 +53,10 @@ static int is_static_frame_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inlin
         *no_inline_reason = "inlining is disabled";
         return 0;
     }
+    if (tc->instance->debugserver) {
+        *no_inline_reason = "inlining not supported when debugging";
+        return 0;
+    }
 
     /* Check frame is not marked as not being allowed to be inlined. */
     if (target_sf->body.no_inline) {
@@ -87,10 +91,12 @@ static int is_static_frame_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inlin
  * and returns zero. Optionally builds up usage counts on the graph. */
 static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                                MVMStaticFrame *target_sf, MVMSpeshIns *invoke_ins,
-                               MVMSpeshGraph *ig, char **no_inline_reason) {
+                               MVMSpeshGraph *ig, char **no_inline_reason, MVMOpInfo **no_inline_info) {
     MVMSpeshBB *bb = ig->entry;
     MVMint32 same_hll = target_sf->body.cu->body.hll_config ==
             inliner->sf->body.cu->body.hll_config;
+    if (no_inline_info)
+        *no_inline_info = NULL;
     while (bb) {
         MVMSpeshIns *ins = bb->first_ins;
         while (ins) {
@@ -101,12 +107,16 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
              * which case we're done. */
             if (!is_phi && ins->info->no_inline) {
                 *no_inline_reason = "target has a :noinline instruction";
+                if (no_inline_info)
+                    *no_inline_info = ins->info;
                 return 0;
             }
 
             /* XXX TODO: See if the workaround below is still required */
             if (opcode == MVM_OP_throwpayloadlexcaller && tc->instance->profiling) {
                 *no_inline_reason = "target has throwpayloadlexcaller, which currently causes problems when profiling is on";
+                if (no_inline_info)
+                    *no_inline_info = ins->info;
                 return 0;
             }
 
@@ -114,6 +124,8 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
              * cannot inline. */
             if (!same_hll && ins->info->uses_hll) {
                 *no_inline_reason = "target has a :useshll instruction and HLLs are different";
+                if (no_inline_info)
+                    *no_inline_info = ins->info;
                 return 0;
             }
 
@@ -133,6 +145,8 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
             if (opcode == MVM_OP_bindlex) {
                 if (ins->operands[0].lex.outers > 0) {
                     *no_inline_reason = "target has bind to outer lexical";
+                    if (no_inline_info)
+                        *no_inline_info = ins->info;
                     return 0;
                 }
             }
@@ -164,6 +178,39 @@ static int is_graph_inlineable(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     return 1;
 }
 
+/* Gets the effective size for inlining considerations of a specialization,
+ * which is its code size minus the code size of its inlines. */
+MVMint32 get_effective_size(MVMThreadContext *tc, MVMSpeshCandidate *cand) {
+    MVMint32 result = cand->bytecode_size;
+    MVMuint32 i;
+    for (i = 0; i < cand->num_inlines; i++)
+        result -= cand->inlines[i].bytecode_size;
+    if (result < 0)
+        result = 0;
+    return (MVMuint32)result;
+}
+
+/* Add deopt usage info to the inlinee. */
+void add_deopt_usages(MVMThreadContext *tc, MVMSpeshGraph *g, MVMint32 *deopt_usage_info,
+                      MVMSpeshIns **deopt_usage_ins) {
+    MVMuint32 usage_idx = 0;
+    MVMuint32 ins_idx = 0;
+    while (deopt_usage_info[usage_idx] != -1) {
+        MVMSpeshIns *ins = deopt_usage_ins[ins_idx++];
+        MVMint32 count = deopt_usage_info[usage_idx + 1];
+        MVMint32 i;
+        usage_idx += 2;
+        for (i = 0; i < count; i++) {
+            MVMint32 deopt_idx = deopt_usage_info[usage_idx++];
+            MVMSpeshFacts *facts = MVM_spesh_get_facts(tc, g, ins->operands[0]);
+            MVMSpeshDeoptUseEntry *entry = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshDeoptUseEntry));
+            entry->deopt_idx = deopt_idx;
+            entry->next = facts->usage.deopt_users;
+            facts->usage.deopt_users = entry;
+        }
+    }
+}
+
 /* Sees if it will be possible to inline the target code ref, given we could
  * already identify a spesh candidate. Returns NULL if no inlining is possible
  * or a graph ready to be merged if it will be possible. */
@@ -171,12 +218,15 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
                                                MVMStaticFrame *target_sf,
                                                MVMSpeshCandidate *cand,
                                                MVMSpeshIns *invoke_ins,
-                                               char **no_inline_reason) {
+                                               char **no_inline_reason,
+                                               MVMuint32 *effective_size,
+                                               MVMOpInfo **no_inline_info) {
     MVMSpeshGraph *ig;
+    MVMSpeshIns **deopt_usage_ins = NULL;
 
     /* Check bytecode size is within the inline limit. */
-    if (cand->bytecode_size > MVM_SPESH_MAX_INLINE_SIZE &&
-            target_sf->body.bytecode_size > MVM_SPESH_MAX_INLINE_SIZE) {
+    *effective_size = get_effective_size(tc, cand);
+    if (*effective_size > MVM_SPESH_MAX_INLINE_SIZE) {
         *no_inline_reason = "bytecode is too large to inline";
         return NULL;
     }
@@ -187,13 +237,14 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
 
     /* Build graph from the already-specialized bytecode and check if we can
      * inline the graph. */
-    ig = MVM_spesh_graph_create_from_cand(tc, target_sf, cand, 0);
-    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason)) {
+    ig = MVM_spesh_graph_create_from_cand(tc, target_sf, cand, 0, &deopt_usage_ins);
+    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason, no_inline_info)) {
         /* We can inline it. Do facts discovery, which also sets usage counts.
          * We also need to bump counts for any inline's code_ref_reg to make
          * sure it stays available for deopt. */
         MVMuint32 i;
         MVM_spesh_facts_discover(tc, ig, NULL, 1);
+        add_deopt_usages(tc, ig, cand->deopt_usage_info, deopt_usage_ins);
         for (i = 0; i < ig->num_inlines; i++) {
             /* We can't be very precise about this, because we don't know the
              * SSA version in effect. So bump usages of all version of the
@@ -203,11 +254,12 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
             for (j = 0; j < ig->fact_counts[reg]; j++)
                 MVM_spesh_usages_add_unconditional_deopt_usage(tc, ig, &(ig->facts[reg][j]));
         }
+        MVM_free(deopt_usage_ins);
         return ig;
     }
     else {
         /* Not possible to inline. Clear it up. */
-        MVM_free(ig->spesh_slots);
+        MVM_free(deopt_usage_ins);
         MVM_spesh_graph_destroy(tc, ig);
         return NULL;
     }
@@ -216,7 +268,7 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph(MVMThreadContext *tc, MVMSpeshGra
 /* Tries to get a spesh graph for a particular unspecialized candidate. */
 MVMSpeshGraph * MVM_spesh_inline_try_get_graph_from_unspecialized(MVMThreadContext *tc,
         MVMSpeshGraph *inliner, MVMStaticFrame *target_sf, MVMSpeshIns *invoke_ins,
-        MVMSpeshCallInfo *call_info, char **no_inline_reason) {
+        MVMSpeshCallInfo *call_info, MVMSpeshStatsType *type_tuple, char **no_inline_reason, MVMOpInfo **no_inline_info) {
     MVMSpeshGraph *ig;
 
     /* Cannot inline with flattening args. */
@@ -234,16 +286,18 @@ MVMSpeshGraph * MVM_spesh_inline_try_get_graph_from_unspecialized(MVMThreadConte
      * is inlineable as we can get rid of many :noinline ops (especially in
      * the args specialization). */
     ig = MVM_spesh_graph_create(tc, target_sf, 0, 1);
-    MVM_spesh_args_from_callinfo(tc, ig, call_info);
+    MVM_spesh_args_from_callinfo(tc, ig, call_info, type_tuple);
     MVM_spesh_facts_discover(tc, ig, NULL, 0);
     MVM_spesh_optimize(tc, ig, NULL);
 
     /* See if it's inlineable; clean up if not. */
-    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason)) {
+    if (is_graph_inlineable(tc, inliner, target_sf, invoke_ins, ig, no_inline_reason, no_inline_info)) {
         return ig;
     }
     else {
-        MVM_free(ig->spesh_slots);
+        /* TODO - we spent all this work creating an optimized version.  Maybe
+           we can make use of it by compiling a custom target despite not being
+           inlineable? */
         MVM_spesh_graph_destroy(tc, ig);
         return NULL;
     }
@@ -408,24 +462,51 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                  MVMSpeshGraph *inlinee, MVMStaticFrame *inlinee_sf,
                  MVMSpeshBB *invoke_bb, MVMSpeshIns *invoke_ins,
                  MVMSpeshOperand code_ref_reg,
-                 MVMuint32 *inline_boundary_handler) {
+                 MVMuint32 *inline_boundary_handler, MVMuint16 bytecode_size) {
     MVMSpeshFacts **merged_facts;
     MVMuint16      *merged_fact_counts;
-    MVMint32        i, orig_inlines, total_inlines, orig_deopt_addrs;
+    MVMint32        i, j, orig_inlines, total_inlines, orig_deopt_addrs,
+                    orig_deopt_pea_mat_infos, impl_deopt_idx;
     MVMuint32       total_handlers = inliner->num_handlers + inlinee->num_handlers + 1;
     MVMSpeshBB     *inlinee_first_bb = NULL, *inlinee_last_bb = NULL;
     MVMuint8        may_cause_deopt = 0;
+    MVMSpeshBB *bb;
+    MVM_VECTOR_DECL(MVMSpeshOperand, regs_for_deopt);
 
-    /* If the inliner and inlinee are from different compilation units, we
-     * potentially have to fix up extra things. */
+    /* If the inliner and inlinee are from different compilation units or
+     * HLLs, we potentially have to fix up extra things. */
     MVMint32 same_comp_unit = inliner->sf->body.cu == inlinee->sf->body.cu;
-
     MVMint32 same_hll = same_comp_unit || inliner->sf->body.cu->body.hll_config ==
             inlinee_sf->body.cu->body.hll_config;
 
+    /* Gather all of the registers that have a deopt usage at the point of
+     * the invoke. We'll need to distribute deopt usages inside of the
+     * inline to them for correctness of later analyses that use the deopt
+     * usage information. */
+    impl_deopt_idx = return_deopt_idx(tc, invoke_ins);
+    MVM_VECTOR_INIT(regs_for_deopt, 0);
+    for (i = 0; i < inliner->sf->body.num_locals; i++) {
+        MVMint32 vers = inliner->fact_counts[i];
+        for (j = 0; j < vers; j++) {
+            MVMSpeshDeoptUseEntry *due = inliner->facts[i][j].usage.deopt_users;
+            while (due) {
+                if (due->deopt_idx == impl_deopt_idx) {
+                    MVMSpeshOperand o;
+                    o.reg.orig = i;
+                    o.reg.i = j;
+                    MVM_VECTOR_PUSH(regs_for_deopt, o);
+                }
+                due = due->next;
+            }
+        }
+    }
+
     /* Renumber the locals, lexicals, and basic blocks of the inlinee; also
-     * re-write any indexes in annotations that need it. */
-    MVMSpeshBB *bb = inlinee->entry;
+     * re-write any indexes in annotations that need it. Since in this pass
+     * we identify all deopt points, we also take the opportunity to add
+     * deopt users to all registers that have deopt usages thanks to the
+     * invoke. */
+    bb = inlinee->entry;
     while (bb) {
         MVMSpeshIns *ins = bb->first_ins;
         while (ins) {
@@ -438,6 +519,9 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                 case MVM_SPESH_ANN_DEOPT_INLINE:
                 case MVM_SPESH_ANN_DEOPT_OSR:
                     ann->data.deopt_idx += inliner->num_deopt_addrs;
+                    for (i = 0; i < MVM_VECTOR_ELEMS(regs_for_deopt); i++)
+                        MVM_spesh_usages_add_deopt_usage_by_reg(tc, inliner,
+                                regs_for_deopt[i], ann->data.deopt_idx);
                     break;
                 case MVM_SPESH_ANN_INLINE_START:
                 case MVM_SPESH_ANN_INLINE_END:
@@ -474,6 +558,14 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                     rewrite_outer_lookup(tc, inliner, ins, inliner->num_locals,
                         MVM_OP_sp_getlexvia_ins, code_ref_reg);
             }
+            else if (opcode == MVM_OP_takedispatcher) {
+                /* If the inlining code doesn't set a dispatcher, takedispatcher
+                 * in the inlinee can be assumed to always return null. */
+                if (!inliner->sets_dispatcher && !inlinee->sets_dispatcher) {
+                    ins->info = MVM_op_get_op(MVM_OP_null);
+                    ins->operands[0].reg.orig += inliner->num_locals;
+                }
+            }
             else {
                 if (!same_comp_unit) {
                     if (ins->info->opcode == MVM_OP_const_s) {
@@ -529,7 +621,20 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
             inlinee_last_bb = bb;
         bb = bb->linear_next;
     }
+    MVM_VECTOR_DESTROY(regs_for_deopt);
 
+    /* If we saw that we may deopt anywhere, mark the basic blocks of the
+     * inline as maybe deopting; this prevents us doing too aggressive
+     * optimizations within them. */
+    if (may_cause_deopt) {
+        bb = inlinee->entry;
+        while (bb) {
+            bb->inlined_may_cause_deopt = 1;
+            bb = bb->linear_next;
+        }
+    }
+
+    /* Link inlinee BBs into the linear next chain. */
     bb = invoke_bb->linear_next;
     invoke_bb->linear_next = inlinee_first_bb = inlinee->entry->linear_next;
     inlinee_last_bb->linear_next = bb;
@@ -547,7 +652,42 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         MVM_spesh_manipulate_add_successor(tc, inliner, inliner->entry, move);
     }
 
-    /* Merge facts. */
+    /* Merge materialization deopt info (for scalar-replaced entities). */
+    orig_deopt_pea_mat_infos = MVM_VECTOR_ELEMS(inliner->deopt_pea.materialize_info);
+    for (i = 0; i < MVM_VECTOR_ELEMS(inlinee->deopt_pea.materialize_info); i++) {
+        MVMSpeshPEAMaterializeInfo mi_orig = inlinee->deopt_pea.materialize_info[i];
+        MVMSpeshPEAMaterializeInfo mi_new;
+        mi_new.stable_sslot = mi_orig.stable_sslot + inliner->num_spesh_slots;
+        mi_new.num_attr_regs = mi_orig.num_attr_regs;
+        if (mi_new.num_attr_regs) {
+            mi_new.attr_regs = MVM_malloc(mi_new.num_attr_regs * sizeof(MVMuint16));
+            for (j = 0; j < mi_new.num_attr_regs; j++)
+                mi_new.attr_regs[j] = mi_orig.attr_regs[j] + inliner->num_locals;
+        }
+        else {
+            mi_new.attr_regs = NULL;
+        }
+        MVM_VECTOR_PUSH(inliner->deopt_pea.materialize_info, mi_new);
+    }
+    for (i = 0; i < MVM_VECTOR_ELEMS(inlinee->deopt_pea.deopt_point); i++) {
+        MVMSpeshPEADeoptPoint dp_orig = inlinee->deopt_pea.deopt_point[i];
+        MVMSpeshPEADeoptPoint dp_new;
+        dp_new.deopt_point_idx = dp_orig.deopt_point_idx + inliner->num_deopt_addrs;
+        dp_new.materialize_info_idx = dp_orig.materialize_info_idx + orig_deopt_pea_mat_infos;
+        dp_new.target_reg = dp_orig.target_reg + inliner->num_locals;
+        MVM_VECTOR_PUSH(inliner->deopt_pea.deopt_point, dp_new);
+    }
+
+    /* Merge facts, fixing up deopt indexes in usage chains. */
+    for (i = 0; i < inlinee->num_locals; i++) {
+        for (j = 0; j < inlinee->fact_counts[i]; j++) {
+            MVMSpeshDeoptUseEntry *due = inlinee->facts[i][j].usage.deopt_users;
+            while (due) {
+                due->deopt_idx += inliner->num_deopt_addrs;
+                due = due->next;
+            }
+        }
+    }
     merged_facts = MVM_spesh_alloc(tc, inliner,
         (inliner->num_locals + inlinee->num_locals) * sizeof(MVMSpeshFacts *));
     memcpy(merged_facts, inliner->facts,
@@ -616,10 +756,10 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
         inliner->inlines[i].locals_start += inliner->num_locals;
         inliner->inlines[i].lexicals_start += inliner->num_lexicals;
         inliner->inlines[i].return_deopt_idx += orig_deopt_addrs;
+        inliner->inlines[i].bytecode_size = 0;
     }
     inliner->inlines[total_inlines - 1].sf             = inlinee_sf;
     inliner->inlines[total_inlines - 1].code_ref_reg   = code_ref_reg.reg.orig;
-    inliner->inlines[total_inlines - 1].g              = inlinee;
     inliner->inlines[total_inlines - 1].locals_start   = inliner->num_locals;
     inliner->inlines[total_inlines - 1].lexicals_start = inliner->num_lexicals;
     inliner->inlines[total_inlines - 1].num_locals     = inlinee->num_locals;
@@ -651,6 +791,7 @@ MVMSpeshBB * merge_graph(MVMThreadContext *tc, MVMSpeshGraph *inliner,
     inliner->inlines[total_inlines - 1].deopt_named_used_bit_field =
         inlinee->deopt_named_used_bit_field;
     inliner->inlines[total_inlines - 1].may_cause_deopt = may_cause_deopt;
+    inliner->inlines[total_inlines - 1].bytecode_size   = bytecode_size;
     inliner->num_inlines = total_inlines;
 
     /* Create/update per-specialization local and lexical type maps. */
@@ -788,12 +929,14 @@ static void tweak_succ(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *bb,
 /* Finds return instructions and re-writes them into gotos, doing any needed
  * boxing or unboxing. */
 static void return_to_set(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshIns *return_ins, MVMSpeshOperand target) {
+    MVMSpeshOperand ver_target = MVM_spesh_manipulate_new_version(tc, g, target.reg.orig);
     MVMSpeshOperand *operands = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
-    operands[0]               = target;
+    operands[0]               = ver_target;
     operands[1]               = return_ins->operands[0];
     return_ins->info          = MVM_op_get_op(MVM_OP_set);
     return_ins->operands      = operands;
-    MVM_spesh_get_facts(tc, g, target)->writer = return_ins;
+    MVM_spesh_get_facts(tc, g, ver_target)->writer = return_ins;
+    MVM_spesh_copy_facts(tc, g, operands[0], operands[1]);
 }
 
 static void return_to_box(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *return_bb,
@@ -802,15 +945,16 @@ static void return_to_box(MVMThreadContext *tc, MVMSpeshGraph *g, MVMSpeshBB *re
     MVMSpeshOperand type_temp     = MVM_spesh_manipulate_get_temp_reg(tc, g, MVM_reg_obj);
 
     /* Create and insert boxing instruction after current return instruction. */
+    MVMSpeshOperand ver_target = MVM_spesh_manipulate_new_version(tc, g, target.reg.orig);
     MVMSpeshIns      *box_ins     = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
     MVMSpeshOperand *box_operands = MVM_spesh_alloc(tc, g, 3 * sizeof(MVMSpeshOperand));
     box_ins->info                 = MVM_op_get_op(box_op);
     box_ins->operands             = box_operands;
-    box_operands[0]               = target;
+    box_operands[0]               = ver_target;
     box_operands[1]               = return_ins->operands[0];
     box_operands[2]               = type_temp;
     MVM_spesh_manipulate_insert_ins(tc, return_bb, return_ins, box_ins);
-    MVM_spesh_get_facts(tc, g, target)->writer = box_ins;
+    MVM_spesh_get_facts(tc, g, ver_target)->writer = box_ins;
     MVM_spesh_usages_add_by_reg(tc, g, box_operands[1], box_ins);
     MVM_spesh_usages_add_by_reg(tc, g, box_operands[2], box_ins);
 
@@ -901,9 +1045,15 @@ static void rewrite_obj_return(MVMThreadContext *tc, MVMSpeshGraph *g,
 static void rewrite_returns(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                      MVMSpeshGraph *inlinee, MVMSpeshBB *invoke_bb,
                      MVMSpeshIns *invoke_ins, MVMSpeshBB *inlinee_last_bb) {
-    /* Locate return instructions. */
+    /* Locate return instructions and rewrite them. For each non-void return,
+     * given the invoke instruction was itself non-void, we generate a new SSA
+     * version of the target register. We then insert a PHI that merges those
+     * versions. */
     MVMSpeshBB *bb = inlinee->entry;
-    MVMuint32 saw_return = 0;
+    MVMint32 initial_last_result_version = invoke_ins->info->opcode != MVM_OP_invoke_v
+        ? inliner->fact_counts[invoke_ins->operands[0].reg.orig]
+        : -1;
+    MVMint32 saw_return = 0;
     while (bb) {
         MVMSpeshIns *ins = bb->first_ins;
         while (ins) {
@@ -952,7 +1102,28 @@ static void rewrite_returns(MVMThreadContext *tc, MVMSpeshGraph *inliner,
             }
             ins = ins->next;
         }
-        if (bb == inlinee_last_bb) break;
+        if (bb == inlinee_last_bb) {
+            MVMint32 final_last_result_version = invoke_ins->info->opcode != MVM_OP_invoke_v
+                ? inliner->fact_counts[invoke_ins->operands[0].reg.orig]
+                : -1;
+            if (final_last_result_version != initial_last_result_version) {
+                /* Produced one or more return results; need a PHI. */
+                MVMuint32 num_rets = final_last_result_version - initial_last_result_version;
+                MVMuint32 i;
+                MVMSpeshIns *phi = MVM_spesh_alloc(tc, inliner, sizeof(MVMSpeshIns));
+                phi->info = get_phi(tc, inliner, num_rets + 1);
+                phi->operands = MVM_spesh_alloc(tc, inliner, (1 + num_rets) * sizeof(MVMSpeshOperand));
+                phi->operands[0] = invoke_ins->operands[0];
+                MVM_spesh_get_facts(tc, inliner, phi->operands[0])->writer = phi;
+                for (i = 0; i < num_rets; i++) {
+                    phi->operands[i + 1].reg.orig = invoke_ins->operands[0].reg.orig;
+                    phi->operands[i + 1].reg.i = initial_last_result_version + i;
+                    MVM_spesh_usages_add_by_reg(tc, inliner, phi->operands[i + 1], phi);
+                }
+                MVM_spesh_manipulate_insert_ins(tc, bb->linear_next, NULL, phi);
+            }
+            break;
+        }
         bb = bb->linear_next;
     }
 }
@@ -1121,13 +1292,13 @@ void MVM_spesh_inline(MVMThreadContext *tc, MVMSpeshGraph *inliner,
                       MVMSpeshCallInfo *call_info, MVMSpeshBB *invoke_bb,
                       MVMSpeshIns *invoke_ins, MVMSpeshGraph *inlinee,
                       MVMStaticFrame *inlinee_sf, MVMSpeshOperand code_ref_reg,
-                      MVMuint32 proxy_deopt_idx) {
+                      MVMuint32 proxy_deopt_idx, MVMuint16 bytecode_size) {
     MVMSpeshIns *first_ins;
 
     /* Merge inlinee's graph into the inliner. */
     MVMuint32 inline_boundary_handler;
     MVMSpeshBB *inlinee_last_bb = merge_graph(tc, inliner, inlinee, inlinee_sf,
-        invoke_bb, invoke_ins, code_ref_reg, &inline_boundary_handler);
+        invoke_bb, invoke_ins, code_ref_reg, &inline_boundary_handler, bytecode_size);
 
     /* If we're profiling, note it's an inline. */
     first_ins = find_first_instruction(tc, inlinee);
@@ -1166,4 +1337,9 @@ void MVM_spesh_inline(MVMThreadContext *tc, MVMSpeshGraph *inliner,
      * for the sake of deopt. */
     if (inliner->inlines[inliner->num_inlines - 1].may_cause_deopt)
         MVM_spesh_usages_retain_deopt_index(tc, inliner, proxy_deopt_idx);
+
+    /* Claim ownership of inlinee memory */
+    MVM_region_merge(tc, &inliner->region_alloc, &inlinee->region_alloc);
+    /* Destroy the inlinee graph */
+    MVM_spesh_graph_destroy(tc, inlinee);
 }
