@@ -59,8 +59,6 @@ MVMuint32 MVM_string_utf16_decodestream_main(MVMThreadContext *tc, MVMDecodeStre
     MVMint32 last_accept_pos, last_was_cr;
     MVMuint32 reached_stopper;
     int low, high;
-    MVMint32  pos = cur_bytes == ds->bytes_head ? ds->bytes_head_pos : 0;
-    MVMuint8 *bytes = (unsigned char *)cur_bytes->bytes;
     /* Set to 1 to remove the BOM even when big endian or little endian are
      * explicitly specified. */
     int remove_bom = 0;
@@ -187,15 +185,16 @@ MVMuint32 MVM_string_utf16_decodestream_main(MVMThreadContext *tc, MVMDecodeStre
 static MVMString * MVM_string_utf16_decode_main(MVMThreadContext *tc,
         const MVMObject *result_type, MVMuint8 *utf16_chars, size_t bytes, int endianess);
 MVMString * MVM_string_utf16be_decode(MVMThreadContext *tc,
-        const MVMObject *result_type, MVMuint8 *utf16_chars, size_t bytes) {
-    return MVM_string_utf16_decode_main(tc, result_type, utf16_chars, bytes, UTF16_DECODE_BIG_ENDIAN);
+        const MVMObject *result_type, char *utf16_chars, size_t bytes) {
+    return MVM_string_utf16_decode_main(tc, result_type, (MVMuint8*)utf16_chars, bytes, UTF16_DECODE_BIG_ENDIAN);
 }
 MVMString * MVM_string_utf16le_decode(MVMThreadContext *tc,
-            const MVMObject *result_type, MVMuint8 *utf16_chars, size_t bytes) {
-    return MVM_string_utf16_decode_main(tc, result_type, utf16_chars, bytes, UTF16_DECODE_LITTLE_ENDIAN);
+            const MVMObject *result_type, char *utf16_chars, size_t bytes) {
+    return MVM_string_utf16_decode_main(tc, result_type, (MVMuint8*)utf16_chars, bytes, UTF16_DECODE_LITTLE_ENDIAN);
 }
 MVMString * MVM_string_utf16_decode(MVMThreadContext *tc,
-            const MVMObject *result_type, MVMuint8 *utf16_chars, size_t bytes) {
+            const MVMObject *result_type, char *utf16_chars_in, size_t bytes) {
+    MVMuint8 *utf16_chars = (MVMuint8*)utf16_chars_in;
 #ifdef MVM_BIGENDIAN
     int mode = UTF16_DECODE_BIG_ENDIAN;
 #else
@@ -214,7 +213,7 @@ MVMString * MVM_string_utf16_decode(MVMThreadContext *tc,
             bytes -= 2;
         }
     }
-    return MVM_string_utf16_decode_main(tc, result_type, utf16_chars, bytes, mode);
+    return MVM_string_utf16_decode_main(tc, result_type, (MVMuint8*)utf16_chars, bytes, mode);
 }
 /* Decodes the specified number of bytes of utf16 into an NFG string, creating
  * a result of the specified type. The type must have the MVMString REPR. */
@@ -399,7 +398,7 @@ char * MVM_string_utf16_encode_substr_main(MVMThreadContext *tc, MVMString *str,
     if (!output_size)
         output_size = &scratch_space;
     *output_size = (char *)result_pos - (char *)result;
-    result = MVM_realloc(result, *output_size);
+    result = MVM_realloc(result, *output_size + 2);
     MVM_free(repl_bytes);
     return (char *)result;
 }
