@@ -125,7 +125,6 @@ typedef struct MVMSemaphoreBody MVMSemaphoreBody;
 typedef struct MVMConcBlockingQueue MVMConcBlockingQueue;
 typedef struct MVMConcBlockingQueueBody MVMConcBlockingQueueBody;
 typedef struct MVMConcBlockingQueueNode MVMConcBlockingQueueNode;
-typedef struct MVMConcBlockingQueueLocks MVMConcBlockingQueueLocks;
 typedef struct MVMObject MVMObject;
 typedef struct MVMObjectId MVMObjectId;
 typedef struct MVMObjectStooge MVMObjectStooge;
@@ -202,10 +201,16 @@ typedef struct MVMSpeshUsages MVMSpeshUsages;
 typedef struct MVMSpeshUseChainEntry MVMSpeshUseChainEntry;
 typedef struct MVMSpeshDeoptUseEntry MVMSpeshDeoptUseEntry;
 typedef struct MVMSpeshFrameWalker MVMSpeshFrameWalker;
+typedef struct MVMSpeshPEAInfo MVMSpeshPEAInfo;
+typedef struct MVMSpeshPEAAllocation MVMSpeshPEAAllocation;
+typedef struct MVMSpeshPEADeopt MVMSpeshPEADeopt;
+typedef struct MVMSpeshPEAMaterializeInfo MVMSpeshPEAMaterializeInfo;
+typedef struct MVMSpeshPEADeoptPoint MVMSpeshPEADeoptPoint;
 typedef struct MVMSTable MVMSTable;
 typedef struct MVMStaticFrame MVMStaticFrame;
 typedef struct MVMStaticFrameBody MVMStaticFrameBody;
 typedef struct MVMStaticFrameInstrumentation MVMStaticFrameInstrumentation;
+typedef struct MVMStaticFrameDebugLocal MVMStaticFrameDebugLocal;
 typedef struct MVMStaticFrameSpesh MVMStaticFrameSpesh;
 typedef struct MVMStaticFrameSpeshBody MVMStaticFrameSpeshBody;
 typedef struct MVMStorageSpec MVMStorageSpec;
